@@ -14,6 +14,7 @@
 #    under the License.
 
 import collections
+import functools
 import os
 import threading
 
@@ -31,7 +32,8 @@ import oslo_messaging
 from oslo_service import loopingcall
 from oslo_utils import fileutils
 from oslo_utils import importutils
-import six
+from oslo_utils import netutils
+from oslo_utils import timeutils
 
 from neutron._i18n import _
 from neutron.agent.common import resource_processing_queue as queue
@@ -39,7 +41,6 @@ from neutron.agent.linux import dhcp
 from neutron.agent.linux import external_process
 from neutron.agent.metadata import driver as metadata_driver
 from neutron.agent import rpc as agent_rpc
-from neutron.common import constants as n_const
 from neutron.common import utils
 from neutron import manager
 
@@ -50,11 +51,14 @@ DEFAULT_PRIORITY = 255
 
 DHCP_PROCESS_GREENLET_MAX = 32
 DHCP_PROCESS_GREENLET_MIN = 8
+DELETED_PORT_MAX_AGE = 86400
+
+DHCP_READY_PORTS_SYNC_MAX = 64
 
 
 def _sync_lock(f):
     """Decorator to block all operations for a global sync call."""
-    @six.wraps(f)
+    @functools.wraps(f)
     def wrapped(*args, **kwargs):
         with _SYNC_STATE_LOCK.write_lock():
             return f(*args, **kwargs)
@@ -63,7 +67,7 @@ def _sync_lock(f):
 
 def _wait_if_syncing(f):
     """Decorator to wait if any sync operations are in progress."""
-    @six.wraps(f)
+    @functools.wraps(f)
     def wrapped(*args, **kwargs):
         with _SYNC_STATE_LOCK.read_lock():
             return f(*args, **kwargs)
@@ -85,6 +89,7 @@ class DhcpAgent(manager.Manager):
         super(DhcpAgent, self).__init__(host=host)
         self.needs_resync_reasons = collections.defaultdict(list)
         self.dhcp_ready_ports = set()
+        self.dhcp_prio_ready_ports = set()
         self.conf = conf or cfg.CONF
         # If 'resync_throttle' is configured more than 'resync_interval' by
         # mistake, raise exception and log with message.
@@ -112,6 +117,20 @@ class DhcpAgent(manager.Manager):
         self._pool_size = DHCP_PROCESS_GREENLET_MIN
         self._pool = eventlet.GreenPool(size=self._pool_size)
         self._queue = queue.ResourceProcessingQueue()
+        self._network_bulk_allocations = {}
+        # Each dhcp-agent restart should trigger a restart of all
+        # metadata-proxies too. This way we can ensure that changes in
+        # the metadata-proxy config we generate will be applied soon
+        # after a new version of dhcp-agent is started. This makes
+        # the metadata service transiently offline. However similar
+        # metadata-proxy restarts were always done by l3-agent so people
+        # can apparently live with short metadata outages. We only stop
+        # the process here and let the process monitor restart it,
+        # first because it knows everything about how to restart it,
+        # second because (unless we temporarily disable the monitor too)
+        # we could race with the monitor restarting the process. See also
+        # method update_isolated_metadata_proxy().
+        self.restarted_metadata_proxy_set = set()
 
     def init_host(self):
         self.sync_state()
@@ -141,11 +160,27 @@ class DhcpAgent(manager.Manager):
         self.periodic_resync()
         self.start_ready_ports_loop()
         eventlet.spawn_n(self._process_loop)
+        if self.conf.bulk_reload_interval:
+            eventlet.spawn_n(self._reload_bulk_allocations)
+
+    def _reload_bulk_allocations(self):
+        while True:
+            for network_id in self._network_bulk_allocations.keys():
+                network = self.cache.get_network_by_id(network_id)
+                self.call_driver('bulk_reload_allocations', network)
+                del self._network_bulk_allocations[network_id]
+            eventlet.greenthread.sleep(self.conf.bulk_reload_interval)
 
     def call_driver(self, action, network, **action_kwargs):
         """Invoke an action on a DHCP driver instance."""
         LOG.debug('Calling driver for network: %(net)s action: %(action)s',
                   {'net': network.id, 'action': action})
+        if self.conf.bulk_reload_interval and action == 'reload_allocations':
+            LOG.debug("Call deferred to bulk load")
+            self._network_bulk_allocations[network.id] = True
+            return True
+        if action == 'bulk_reload_allocations':
+            action = 'reload_allocations'
         try:
             # the Driver expects something that is duck typed similar to
             # the base models.
@@ -154,8 +189,11 @@ class DhcpAgent(manager.Manager):
                                           self._process_monitor,
                                           self.dhcp_version,
                                           self.plugin_rpc)
-            getattr(driver, action)(**action_kwargs)
-            return True
+            rv = getattr(driver, action)(**action_kwargs)
+            if action == 'get_metadata_bind_interface':
+                return rv
+            else:
+                return True
         except exceptions.Conflict:
             # No need to resync here, the agent will receive the event related
             # to a status update for the network
@@ -240,13 +278,23 @@ class DhcpAgent(manager.Manager):
     def _dhcp_ready_ports_loop(self):
         """Notifies the server of any ports that had reservations setup."""
         while True:
-            # this is just watching a set so we can do it really frequently
+            # this is just watching sets so we can do it really frequently
             eventlet.sleep(0.1)
-            if self.dhcp_ready_ports:
-                ports_to_send = self.dhcp_ready_ports
-                self.dhcp_ready_ports = set()
+            prio_ports_to_send = set()
+            ports_to_send = set()
+            for port_count in range(min(len(self.dhcp_prio_ready_ports) +
+                                        len(self.dhcp_ready_ports),
+                                        DHCP_READY_PORTS_SYNC_MAX)):
+                if self.dhcp_prio_ready_ports:
+                    prio_ports_to_send.add(self.dhcp_prio_ready_ports.pop())
+                    continue
+                ports_to_send.add(self.dhcp_ready_ports.pop())
+            if prio_ports_to_send or ports_to_send:
                 try:
-                    self.plugin_rpc.dhcp_ready_on_ports(ports_to_send)
+                    self.plugin_rpc.dhcp_ready_on_ports(prio_ports_to_send |
+                                                        ports_to_send)
+                    LOG.info("DHCP configuration for ports %s is completed",
+                             prio_ports_to_send | ports_to_send)
                     continue
                 except oslo_messaging.MessagingTimeout:
                     LOG.error("Timeout notifying server of ports ready. "
@@ -255,6 +303,7 @@ class DhcpAgent(manager.Manager):
                     LOG.exception("Failure notifying DHCP server of "
                                   "ready DHCP ports. Will retry on next "
                                   "iteration.")
+                self.dhcp_prio_ready_ports |= prio_ports_to_send
                 self.dhcp_ready_ports |= ports_to_send
 
     def start_ready_ports_loop(self):
@@ -536,9 +585,10 @@ class DhcpAgent(manager.Manager):
         network = self.cache.get_network_by_id(updated_port.network_id)
         if not network:
             return
-        self.reload_allocations(updated_port, network)
+        # treat update as a create event
+        self.reload_allocations(updated_port, network, prio=True)
 
-    def reload_allocations(self, port, network):
+    def reload_allocations(self, port, network, prio=False):
         LOG.info("Trigger reload_allocations for port %s", port)
         driver_action = 'reload_allocations'
         if self._is_port_on_this_agent(port):
@@ -564,7 +614,10 @@ class DhcpAgent(manager.Manager):
                 driver_action = 'restart'
         self.cache.put_port(port)
         self.call_driver(driver_action, network)
-        self.dhcp_ready_ports.add(port.id)
+        if prio:
+            self.dhcp_prio_ready_ports.add(port.id)
+        else:
+            self.dhcp_ready_ports.add(port.id)
         self.update_isolated_metadata_proxy(network)
 
     def _is_port_on_this_agent(self, port):
@@ -601,7 +654,7 @@ class DhcpAgent(manager.Manager):
                                      "DHCP cache is out of sync",
                                      created_port.network_id)
                 return
-        self.reload_allocations(created_port, network)
+        self.reload_allocations(created_port, network, prio=True)
 
     def port_delete_end(self, context, payload):
         """Handle the port.delete.end notification event."""
@@ -622,7 +675,7 @@ class DhcpAgent(manager.Manager):
             return
         port_id = payload['port_id']
         port = self.cache.get_port_by_id(port_id)
-        self.cache.deleted_ports.add(port_id)
+        self.cache.add_to_deleted_ports(port_id)
         if not port:
             return
         network = self.cache.get_network_by_id(port.network_id)
@@ -642,11 +695,15 @@ class DhcpAgent(manager.Manager):
 
         According to return from driver class, spawn or kill the metadata
         proxy process. Spawn an existing metadata proxy or kill a nonexistent
-        metadata proxy will just silently return.
+        metadata proxy will just silently return. Spawning an existing
+        metadata proxy restarts it once after each dhcp-agent start.
         """
         should_enable_metadata = self.dhcp_driver_cls.should_enable_metadata(
             self.conf, network)
         if should_enable_metadata:
+            if network.id not in self.restarted_metadata_proxy_set:
+                self.disable_isolated_metadata_proxy(network)
+                self.restarted_metadata_proxy_set.add(network.id)
             self.enable_isolated_metadata_proxy(network)
         else:
             self.disable_isolated_metadata_proxy(network)
@@ -679,9 +736,33 @@ class DhcpAgent(manager.Manager):
                     self._metadata_routers[network.id] = (
                         router_ports[0].device_id)
 
+        if netutils.is_ipv6_enabled():
+            try:
+                dhcp_ifaces = [
+                    self.call_driver(
+                        'get_metadata_bind_interface', network, port=p)
+                    for p in network.ports
+                    if (p.device_owner == constants.DEVICE_OWNER_DHCP and
+                        p.admin_state_up)
+                ]
+                if len(dhcp_ifaces) == 1:
+                    kwargs['bind_interface'] = dhcp_ifaces[0]
+                    kwargs['bind_address_v6'] = constants.METADATA_V6_IP
+                else:
+                    LOG.error(
+                        'Unexpected number of DHCP interfaces for metadata '
+                        'proxy, expected 1, got %s', len(dhcp_ifaces)
+                    )
+            except AttributeError:
+                LOG.warning(
+                    'Cannot serve metadata on IPv6 because DHCP driver '
+                    'does not implement method '
+                    'get_metadata_bind_interface(): %s',
+                    self.dhcp_driver_class)
+
         metadata_driver.MetadataDriver.spawn_monitored_metadata_proxy(
-            self._process_monitor, network.namespace, dhcp.METADATA_PORT,
-            self.conf, bind_address=dhcp.METADATA_DEFAULT_IP, **kwargs)
+            self._process_monitor, network.namespace, constants.METADATA_PORT,
+            self.conf, bind_address=constants.METADATA_V4_IP, **kwargs)
 
     def disable_isolated_metadata_proxy(self, network):
         if (self.conf.enable_metadata_network and
@@ -711,14 +792,15 @@ class DhcpPluginApi(object):
         1.1 - Added get_active_networks_info, create_dhcp_port,
               and update_dhcp_port methods.
         1.5 - Added dhcp_ready_on_ports
-
+        1.7 - Added get_networks
+        1.8 - Added get_dhcp_port
     """
 
     def __init__(self, topic, host):
         self.host = host
         target = oslo_messaging.Target(
                 topic=topic,
-                namespace=n_const.RPC_NAMESPACE_DHCP_PLUGIN,
+                namespace=constants.RPC_NAMESPACE_DHCP_PLUGIN,
                 version='1.0')
         self.client = n_rpc.get_client(target)
 
@@ -769,12 +851,33 @@ class DhcpPluginApi(object):
                           network_id=network_id, device_id=device_id,
                           host=self.host)
 
+    def get_dhcp_port(self, port_id):
+        """Make a remote process call to retrieve the dhcp port."""
+        cctxt = self.client.prepare(version='1.8')
+        port = cctxt.call(self.context, 'get_dhcp_port', port_id=port_id)
+        if port:
+            return dhcp.DictModel(port)
+
     def dhcp_ready_on_ports(self, port_ids):
         """Notify the server that DHCP is configured for the port."""
         cctxt = self.client.prepare(version='1.5')
         return cctxt.call(self.context, 'dhcp_ready_on_ports',
                           port_ids=port_ids)
 
+    def get_networks(self, filters=None, fields=None):
+        """Get networks.
+
+        :param filters: The filters to apply.
+                        E.g {"id" : ["<uuid of a network>", ...]}
+        :param fields: A list of fields to collect, e.g ["id", "subnets"].
+        :return: A list of NetModel where each object represent a network.
+        """
+
+        cctxt = self.client.prepare(version='1.7')
+        nets = cctxt.call(self.context, 'get_networks', filters=filters,
+                          fields=fields)
+        return [dhcp.NetModel(net) for net in nets]
+
 
 class NetworkCache(object):
     """Agent cache of the current network state."""
@@ -782,13 +885,18 @@ class NetworkCache(object):
         self.cache = {}
         self.subnet_lookup = {}
         self.port_lookup = {}
-        self.deleted_ports = set()
+        self._deleted_ports = set()
+        self._deleted_ports_ts = []
+        self.cleanup_loop = loopingcall.FixedIntervalLoopingCall(
+            self.cleanup_deleted_ports)
+        self.cleanup_loop.start(DELETED_PORT_MAX_AGE,
+                                initial_delay=DELETED_PORT_MAX_AGE)
 
     def is_port_message_stale(self, payload):
         orig = self.get_port_by_id(payload['id']) or {}
         if orig.get('revision_number', 0) > payload.get('revision_number', 0):
             return True
-        if payload['id'] in self.deleted_ports:
+        if payload['id'] in self._deleted_ports:
             return True
         return False
 
@@ -875,6 +983,29 @@ class NetworkCache(object):
                 'subnets': num_subnets,
                 'ports': num_ports}
 
+    def add_to_deleted_ports(self, port_id):
+        if port_id not in self._deleted_ports:
+            self._deleted_ports.add(port_id)
+            self._deleted_ports_ts.append((timeutils.utcnow_ts(), port_id))
+
+    def cleanup_deleted_ports(self):
+        """Cleanup the "self._deleted_ports" set based on the current TS
+
+        The variable "self._deleted_ports_ts" contains a timestamp
+        ordered list of tuples (timestamp, port_id). Every port older than the
+        current timestamp minus "timestamp_delta" will be deleted from
+        "self._deleted_ports" and "self._deleted_ports_ts".
+        """
+        timestamp_min = timeutils.utcnow_ts() - DELETED_PORT_MAX_AGE
+        idx = None
+        for idx, (ts, port_id) in enumerate(self._deleted_ports_ts):
+            if ts > timestamp_min:
+                break
+            self._deleted_ports.remove(port_id)
+
+        if idx:
+            self._deleted_ports_ts = self._deleted_ports_ts[idx:]
+
 
 class DhcpAgentWithStateReport(DhcpAgent):
     def __init__(self, host=None, conf=None):
@@ -882,7 +1013,7 @@ class DhcpAgentWithStateReport(DhcpAgent):
         self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
         self.failed_report_state = False
         self.agent_state = {
-            'binary': 'neutron-dhcp-agent',
+            'binary': constants.AGENT_PROCESS_DHCP,
             'host': host,
             'availability_zone': self.conf.AGENT.availability_zone,
             'topic': topics.DHCP_AGENT,
