@@ -17,14 +17,16 @@ import itertools
 import os
 
 import netaddr
+from neutron_lib import constants
 from neutron_lib import exceptions
+from neutron_lib.utils import file as file_utils
 from oslo_config import cfg
 from oslo_log import log as logging
+from oslo_utils import fileutils
 
-from neutron._i18n import _, _LE
+from neutron._i18n import _
 from neutron.agent.linux import external_process
-from neutron.common import constants
-from neutron.common import utils as common_utils
+from neutron.common import utils
 
 VALID_STATES = ['MASTER', 'BACKUP']
 VALID_AUTH_TYPES = ['AH', 'PASS']
@@ -33,7 +35,8 @@ PRIMARY_VIP_RANGE_SIZE = 24
 KEEPALIVED_SERVICE_NAME = 'keepalived'
 KEEPALIVED_EMAIL_FROM = 'neutron@openstack.local'
 KEEPALIVED_ROUTER_ID = 'neutron'
-GARP_MASTER_DELAY = 60
+GARP_PRIMARY_DELAY = 60
+HEALTH_CHECK_NAME = 'ha_health_check'
 
 LOG = logging.getLogger(__name__)
 
@@ -82,24 +85,28 @@ class InvalidAuthenticationTypeException(exceptions.NeutronException):
 class KeepalivedVipAddress(object):
     """A virtual address entry of a keepalived configuration."""
 
-    def __init__(self, ip_address, interface_name, scope=None):
+    def __init__(self, ip_address, interface_name, scope=None, track=True):
         self.ip_address = ip_address
         self.interface_name = interface_name
         self.scope = scope
+        self.track = track
 
     def __eq__(self, other):
         return (isinstance(other, KeepalivedVipAddress) and
                 self.ip_address == other.ip_address)
 
     def __str__(self):
-        return '[%s, %s, %s]' % (self.ip_address,
-                                 self.interface_name,
-                                 self.scope)
+        return '[%s, %s, %s, %s]' % (self.ip_address,
+                                     self.interface_name,
+                                     self.scope,
+                                     self.track)
 
     def build_config(self):
         result = '%s dev %s' % (self.ip_address, self.interface_name)
         if self.scope:
             result += ' scope %s' % self.scope
+        if cfg.CONF.keepalived_use_no_track and not self.track:
+            result += ' no_track'
         return result
 
 
@@ -121,6 +128,8 @@ class KeepalivedVirtualRoute(object):
             output += ' dev %s' % self.interface_name
         if self.scope:
             output += ' scope %s' % self.scope
+        if cfg.CONF.keepalived_use_no_track:
+            output += ' no_track'
         return output
 
 
@@ -159,7 +168,9 @@ class KeepalivedInstance(object):
     def __init__(self, state, interface, vrouter_id, ha_cidrs,
                  priority=HA_DEFAULT_PRIORITY, advert_int=None,
                  mcast_src_ip=None, nopreempt=False,
-                 garp_master_delay=GARP_MASTER_DELAY):
+                 garp_primary_delay=GARP_PRIMARY_DELAY,
+                 vrrp_health_check_interval=0,
+                 ha_conf_dir=None):
         self.name = 'VR_%s' % vrouter_id
 
         if state not in VALID_STATES:
@@ -172,17 +183,22 @@ class KeepalivedInstance(object):
         self.nopreempt = nopreempt
         self.advert_int = advert_int
         self.mcast_src_ip = mcast_src_ip
-        self.garp_master_delay = garp_master_delay
+        self.garp_primary_delay = garp_primary_delay
         self.track_interfaces = []
         self.vips = []
         self.virtual_routes = KeepalivedInstanceRoutes()
         self.authentication = None
+        self.track_script = None
         self.primary_vip_range = get_free_range(
             parent_range=constants.PRIVATE_CIDR_RANGE,
             excluded_ranges=[constants.METADATA_CIDR,
                              constants.DVR_FIP_LL_CIDR] + ha_cidrs,
             size=PRIMARY_VIP_RANGE_SIZE)
 
+        if vrrp_health_check_interval > 0:
+            self.track_script = KeepalivedTrackScript(
+                vrrp_health_check_interval, ha_conf_dir, self.vrouter_id)
+
     def set_authentication(self, auth_type, password):
         if auth_type not in VALID_AUTH_TYPES:
             raise InvalidAuthenticationTypeException(auth_type=auth_type)
@@ -190,7 +206,8 @@ class KeepalivedInstance(object):
         self.authentication = (auth_type, password)
 
     def add_vip(self, ip_cidr, interface_name, scope):
-        vip = KeepalivedVipAddress(ip_cidr, interface_name, scope)
+        track = interface_name in self.track_interfaces
+        vip = KeepalivedVipAddress(ip_cidr, interface_name, scope, track=track)
         if vip not in self.vips:
             self.vips.append(vip)
         else:
@@ -266,12 +283,19 @@ class KeepalivedInstance(object):
                                ['    }'])
 
     def build_config(self):
-        config = ['vrrp_instance %s {' % self.name,
-                  '    state %s' % self.state,
-                  '    interface %s' % self.interface,
-                  '    virtual_router_id %s' % self.vrouter_id,
-                  '    priority %s' % self.priority,
-                  '    garp_master_delay %s' % self.garp_master_delay]
+        if self.track_script:
+            config = self.track_script.build_config_preamble()
+            self.track_script.routes = self.virtual_routes.gateway_routes
+            self.track_script.vips = self.vips
+        else:
+            config = []
+
+        config.extend(['vrrp_instance %s {' % self.name,
+                       '    state %s' % self.state,
+                       '    interface %s' % self.interface,
+                       '    virtual_router_id %s' % self.vrouter_id,
+                       '    priority %s' % self.priority,
+                       '    garp_master_delay %s' % self.garp_primary_delay])
 
         if self.nopreempt:
             config.append('    nopreempt')
@@ -298,6 +322,9 @@ class KeepalivedInstance(object):
         if len(self.virtual_routes):
             config.extend(self.virtual_routes.build_config())
 
+        if self.track_script:
+            config.extend(self.track_script.build_config())
+
         config.append('}')
 
         return config
@@ -346,13 +373,21 @@ class KeepalivedManager(object):
 
     """
 
-    def __init__(self, resource_id, config, process_monitor, conf_path='/tmp',
-                 namespace=None):
+    def __init__(self, resource_id, config, process_monitor, conf_path,
+                 namespace=None, throttle_restart_value=None):
         self.resource_id = resource_id
         self.config = config
         self.namespace = namespace
         self.process_monitor = process_monitor
         self.conf_path = conf_path
+        # configure throttler for spawn to introduce delay between SIGHUPs,
+        # otherwise keepalived primary may unnecessarily flip to backup
+        if throttle_restart_value is not None:
+            self._throttle_spawn(throttle_restart_value)
+
+    # pylint: disable=method-hidden
+    def _throttle_spawn(self, threshold):
+        self.spawn = utils.throttler(threshold)(self.spawn)
 
     def get_conf_dir(self):
         confs_dir = os.path.abspath(os.path.normpath(self.conf_path))
@@ -362,13 +397,15 @@ class KeepalivedManager(object):
     def get_full_config_file_path(self, filename, ensure_conf_dir=True):
         conf_dir = self.get_conf_dir()
         if ensure_conf_dir:
-            common_utils.ensure_dir(conf_dir)
+            fileutils.ensure_tree(conf_dir, mode=0o755)
         return os.path.join(conf_dir, filename)
 
     def _output_config_file(self):
         config_str = self.config.get_config_str()
+        LOG.debug("Router %s keepalived config: %s",
+                  self.resource_id, config_str)
         config_path = self.get_full_config_file_path('keepalived.conf')
-        common_utils.replace_file(config_path, config_str)
+        file_utils.replace_file(config_path, config_str)
 
         return config_path
 
@@ -378,8 +415,8 @@ class KeepalivedManager(object):
             os.remove(pid_file)
         except OSError as e:
             if e.errno != errno.ENOENT:
-                LOG.error(_LE("Could not delete file %s, keepalived can "
-                              "refuse to start."), pid_file)
+                LOG.error("Could not delete file %s, keepalived can "
+                          "refuse to start.", pid_file)
 
     def get_vrrp_pid_file_name(self, base_pid_file):
         return '%s-vrrp' % base_pid_file
@@ -396,6 +433,10 @@ class KeepalivedManager(object):
     def spawn(self):
         config_path = self._output_config_file()
 
+        for key, instance in self.config.instances.items():
+            if instance.track_script:
+                instance.track_script.write_check_script()
+
         keepalived_pm = self.get_process()
         vrrp_pm = self._get_vrrp_process(
             self.get_vrrp_pid_file_name(keepalived_pm.get_pid_file_name()))
@@ -418,11 +459,18 @@ class KeepalivedManager(object):
         pm = self.get_process()
         pm.disable(sig='15')
 
+    def check_processes(self):
+        keepalived_pm = self.get_process()
+        vrrp_pm = self._get_vrrp_process(
+            self.get_vrrp_pid_file_name(keepalived_pm.get_pid_file_name()))
+        return keepalived_pm.active and vrrp_pm.active
+
     def get_process(self):
         return external_process.ProcessManager(
             cfg.CONF,
             self.resource_id,
             self.namespace,
+            service=KEEPALIVED_SERVICE_NAME,
             pids_path=self.conf_path)
 
     def _get_vrrp_process(self, pid_file):
@@ -449,6 +497,86 @@ class KeepalivedManager(object):
                    '-f', config_path,
                    '-p', pid_file,
                    '-r', self.get_vrrp_pid_file_name(pid_file)]
+            if logging.is_debug_enabled(cfg.CONF):
+                cmd.append('-D')
             return cmd
 
         return callback
+
+
+class KeepalivedTrackScript(KeepalivedConf):
+    """Track script generator for Keepalived"""
+
+    def __init__(self, interval, conf_dir, vr_id):
+        self.interval = interval
+        self.conf_dir = conf_dir
+        self.vr_id = vr_id
+        self.routes = []
+        self.vips = []
+
+    def build_config_preamble(self):
+        config = ['',
+                  'vrrp_script %s_%s {' % (HEALTH_CHECK_NAME, self.vr_id),
+                  '    script "%s"' % self._get_script_location(),
+                  '    interval %s' % self.interval,
+                  '    fall 2',
+                  '    rise 2',
+                  '}',
+                  '']
+
+        return config
+
+    def _is_needed(self):
+        """Check if track script is needed by checking amount of routes.
+
+        :return: True/False
+        """
+        return len(self.routes) > 0
+
+    def build_config(self):
+        if not self._is_needed():
+            return ''
+
+        config = ['    track_script {',
+                  '        %s_%s' % (HEALTH_CHECK_NAME, self.vr_id),
+                  '    }']
+
+        return config
+
+    def build_script(self):
+        return itertools.chain(['#!/bin/bash -eu'],
+                               ['%s' % self._check_ip_assigned()],
+                               ('%s' % self._add_ip_addr(route.nexthop)
+                                for route in self.routes if route.nexthop),
+                               )
+
+    def _add_ip_addr(self, ip_addr):
+        cmd = {
+            4: 'ping',
+            6: 'ping6',
+        }.get(netaddr.IPAddress(ip_addr).version)
+
+        return '%s -c 1 -w 1 %s 1>/dev/null || exit 1' % (cmd, ip_addr)
+
+    def _check_ip_assigned(self):
+        cmd = 'ip a | grep %s || exit 0'
+        return cmd % netaddr.IPNetwork(self.vips[0].ip_address).ip if len(
+            self.vips) else ''
+
+    def _get_script_str(self):
+        """Generates and returns bash script to verify connectivity.
+
+        :return: Bash script code
+        """
+        return '\n'.join(self.build_script())
+
+    def _get_script_location(self):
+        return os.path.join(self.conf_dir,
+                            'ha_check_script_%s.sh' % self.vr_id)
+
+    def write_check_script(self):
+        if not self._is_needed():
+            return
+
+        file_utils.replace_file(
+            self._get_script_location(), self._get_script_str(), 0o520)
