@@ -13,16 +13,24 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import builtins
+import ctypes
+import io
 import os
 
 import eventlet
 from eventlet import tpool
+from neutron_lib import exceptions
 from neutron_lib.utils import helpers
 from oslo_log import log as logging
 from oslo_utils import encodeutils
 
 from neutron._i18n import _
 
+
+if os.name == 'nt':
+    import wmi
+
 LOG = logging.getLogger(__name__)
 
 # subprocess.Popen will spawn two threads consuming stdout/stderr when passing
@@ -31,8 +39,11 @@ LOG = logging.getLogger(__name__)
 subprocess = eventlet.patcher.original('subprocess')
 subprocess.threading = eventlet.patcher.original('threading')
 
+ERROR_KEY_DELETED = 0x03FA
 
-def create_process(cmd, addl_env=None):
+
+def create_process(cmd, run_as_root=False, addl_env=None,
+                   tpool_proxy=True):
     cmd = list(map(str, cmd))
 
     LOG.debug("Running command: %s", cmd)
@@ -48,10 +59,47 @@ def create_process(cmd, addl_env=None):
                 env=env,
                 preexec_fn=None,
                 close_fds=False)
+    if tpool_proxy and eventlet.getcurrent().parent:
+        # If we intend to access the process streams, we need to wrap this
+        # in a tpool proxy object, avoding blocking other greenthreads.
+        #
+        # The 'file' type is not available on Python 3.x.
+        file_type = getattr(builtins, 'file', io.IOBase)
+        obj = tpool.Proxy(obj, autowrap=(file_type, ))
 
     return obj, cmd
 
 
+def _get_wmi_process(pid):
+    if not pid:
+        return None
+
+    try:
+        conn = wmi.WMI()
+        processes = conn.Win32_Process(ProcessId=pid)
+    except wmi.x_wmi as exc:
+        hresult = exc.com_error.hresult
+        err_code = ctypes.c_uint(hresult).value & 0xFFFF
+        if err_code == ERROR_KEY_DELETED:
+            return None
+        raise
+
+    if processes:
+        return processes[0]
+    return None
+
+
+def kill_process(pid, signal, run_as_root=False):
+    """Kill the process with the given pid using the given signal."""
+    process = _get_wmi_process(pid)
+    try:
+        if process:
+            process.Terminate()
+    except Exception:
+        if _get_wmi_process(pid):
+            raise
+
+
 def execute(cmd, process_input=None, addl_env=None,
             check_exit_code=True, return_stderr=False, log_fail_as_error=True,
             extra_ok_codes=None, run_as_root=False, do_decode=True):
@@ -60,7 +108,7 @@ def execute(cmd, process_input=None, addl_env=None,
         _process_input = encodeutils.to_utf8(process_input)
     else:
         _process_input = None
-    obj, cmd = create_process(cmd, addl_env=addl_env)
+    obj, cmd = create_process(cmd, addl_env=addl_env, tpool_proxy=False)
     _stdout, _stderr = avoid_blocking_call(obj.communicate, _process_input)
     obj.stdin.close()
     _stdout = helpers.safe_decode_utf8(_stdout)
@@ -85,7 +133,7 @@ def execute(cmd, process_input=None, addl_env=None,
         LOG.debug(log_msg)
 
     if obj.returncode and check_exit_code:
-        raise RuntimeError(m)
+        raise exceptions.ProcessExecutionError(m, returncode=obj.returncode)
 
     return (_stdout, _stderr) if return_stderr else _stdout
 
@@ -107,3 +155,22 @@ def avoid_blocking_call(f, *args, **kwargs):
         return tpool.execute(f, *args, **kwargs)
     else:
         return f(*args, **kwargs)
+
+
+def get_root_helper_child_pid(pid, expected_cmd, run_as_root=False):
+    # We don't use a root helper on Windows.
+    return str(pid)
+
+
+def process_is_running(pid):
+    """Find if the given PID is running in the system."""
+    return _get_wmi_process(pid) is not None
+
+
+def pid_invoked_with_cmdline(pid, expected_cmd):
+    process = _get_wmi_process(pid)
+    if not process:
+        return False
+
+    command = process.CommandLine
+    return command == " ".join(expected_cmd)
