@@ -32,7 +32,6 @@ from oslo_log import log as logging
 import oslo_messaging
 from oslo_service import service
 from oslo_utils import excutils
-from six import moves
 
 from neutron.agent.linux import bridge_lib
 from neutron.agent.linux import ip_lib
@@ -41,6 +40,7 @@ from neutron.common import config as common_config
 from neutron.common import profiler as setup_profiler
 from neutron.common import utils
 from neutron.conf.agent import common as agent_config
+from neutron.conf import service as service_conf
 from neutron.plugins.ml2.drivers.agent import _agent_manager_base as amb
 from neutron.plugins.ml2.drivers.agent import _common_agent as ca
 from neutron.plugins.ml2.drivers.agent import config as cagt_config  # noqa
@@ -60,7 +60,6 @@ from neutron.plugins.ml2.drivers.linuxbridge.agent \
 
 LOG = logging.getLogger(__name__)
 
-LB_AGENT_BINARY = 'neutron-linuxbridge-agent'
 BRIDGE_NAME_PREFIX = "brq"
 MAX_VLAN_POSTFIX_LEN = 5
 VXLAN_INTERFACE_PREFIX = "vxlan-"
@@ -397,11 +396,12 @@ class LinuxBridgeManager(amb.CommonAgentManagerBase):
         if gateway:
             # Ensure that the gateway can be updated by changing the metric
             metric = 100
-            if 'metric' in gateway:
+            ip_version = utils.get_ip_version(gateway['cidr'])
+            if gateway['metric'] != ip_lib.IP_ROUTE_METRIC_DEFAULT[ip_version]:
                 metric = gateway['metric'] - 1
-            dst_device.route.add_gateway(gateway=gateway['gateway'],
+            dst_device.route.add_gateway(gateway=gateway['via'],
                                          metric=metric)
-            src_device.route.delete_gateway(gateway=gateway['gateway'])
+            src_device.route.delete_gateway(gateway=gateway['via'])
 
         # Remove IP's from interface
         if ips:
@@ -458,7 +458,7 @@ class LinuxBridgeManager(amb.CommonAgentManagerBase):
         # Check if the interface is part of the bridge
         if not bridge_device.owns_interface(interface):
             try:
-                # Check if the interface is not enslaved in another bridge
+                # Check if the interface is attached to another bridge
                 bridge = bridge_lib.BridgeDevice.get_interface_bridge(
                     interface)
                 if bridge:
@@ -699,7 +699,7 @@ class LinuxBridgeManager(amb.CommonAgentManagerBase):
             return False
 
         test_iface = None
-        for seg_id in moves.range(1, constants.MAX_VXLAN_VNI + 1):
+        for seg_id in range(1, constants.MAX_VXLAN_VNI + 1):
             if (ip_lib.device_exists(self.get_vxlan_device_name(seg_id)) or
                     ip_lib.vxlan_in_use(seg_id)):
                 continue
@@ -1030,6 +1030,8 @@ def main():
 
     common_config.setup_logging()
     agent_config.setup_privsep()
+    service_conf.register_service_opts(service_conf.RPC_EXTRA_OPTS, cfg.CONF)
+
     try:
         interface_mappings = helpers.parse_mappings(
             cfg.CONF.LINUX_BRIDGE.physical_interface_mappings)
@@ -1055,8 +1057,8 @@ def main():
     quitting_rpc_timeout = cfg.CONF.AGENT.quitting_rpc_timeout
     agent = ca.CommonAgentLoop(manager, polling_interval, quitting_rpc_timeout,
                                constants.AGENT_TYPE_LINUXBRIDGE,
-                               LB_AGENT_BINARY)
-    setup_profiler.setup("neutron-linuxbridge-agent", cfg.CONF.host)
+                               constants.AGENT_PROCESS_LINUXBRIDGE)
+    setup_profiler.setup(constants.AGENT_PROCESS_LINUXBRIDGE, cfg.CONF.host)
     LOG.info("Agent initialized successfully, now running... ")
     launcher = service.launch(cfg.CONF, agent, restart_method='mutate')
     launcher.wait()
