@@ -14,13 +14,18 @@
 #    under the License.
 
 import collections
+from unittest import mock
 import uuid
 
-import mock
+from neutron_lib import constants as const
+from oslo_config import cfg
+from ovsdbapp.backend.ovs_idl import idlutils
 
 from neutron.agent.common import ovs_lib
 from neutron.agent.linux import ip_lib
 from neutron.common import utils
+from neutron.plugins.ml2.drivers.openvswitch.agent.common import (
+    constants as agent_const)
 from neutron.tests.common import net_helpers
 from neutron.tests.functional.agent.linux import base
 
@@ -85,6 +90,8 @@ class OVSBridgeTestCase(OVSBridgeTestBase):
         self.assertTrue(self.br.port_exists(port_name))
         self.assertEqual('test', self.br.db_get_val('Interface', port_name,
                                                     'external_ids')['test'])
+        self.assertEqual(agent_const.DEAD_VLAN_TAG,
+                         self.br.db_get_val('Port', port_name, 'tag'))
 
     def test_attribute_lifecycle(self):
         (port_name, ofport) = self.create_ovs_port()
@@ -96,6 +103,23 @@ class OVSBridgeTestCase(OVSBridgeTestBase):
         self.assertEqual([], self.ovs.db_get_val('Port', port_name, 'tag'))
         self.assertEqual([], self.br.get_port_tag_dict()[port_name])
 
+    def test_attribute_map_handling(self):
+        (pname, ofport) = self.create_ovs_port()
+        expected = {'a': 'b'}
+        self.ovs.set_db_attribute('Port', pname, 'other_config', expected)
+        self.assertEqual(expected,
+                         self.ovs.db_get_val('Port', pname, 'other_config'))
+        other = {'c': 'd'}
+        expected.update(other)
+        self.ovs.set_db_attribute('Port', pname, 'other_config', other)
+        self.assertEqual(expected,
+                         self.ovs.db_get_val('Port', pname, 'other_config'))
+        other = {'a': 'x'}
+        expected.update(other)
+        self.ovs.set_db_attribute('Port', pname, 'other_config', other)
+        self.assertEqual(expected,
+                         self.ovs.db_get_val('Port', pname, 'other_config'))
+
     def test_get_bridge_external_bridge_id(self):
         self.ovs.set_db_attribute('Bridge', self.br.br_name,
                                   'external_ids',
@@ -134,11 +158,51 @@ class OVSBridgeTestCase(OVSBridgeTestBase):
             self.br.db_get_val('Bridge', self.br.br_name, 'fail_mode'),
             fail_mode)
 
-    def test_set_protocols(self):
-        self.br.set_protocols('OpenFlow10')
+    def test_add_protocols_start_with_one(self):
+        self.br.set_db_attribute('Bridge', self.br.br_name, 'protocols',
+                                 ['OpenFlow10'],
+                                 check_error=True)
+        self.br.add_protocols('OpenFlow13')
         self.assertEqual(
             self.br.db_get_val('Bridge', self.br.br_name, 'protocols'),
-            "OpenFlow10")
+            ['OpenFlow10', 'OpenFlow13'])
+
+    def test_add_protocols_start_with_two_add_two(self):
+        self.br.set_db_attribute('Bridge', self.br.br_name, 'protocols',
+                                 ['OpenFlow10', 'OpenFlow12'],
+                                 check_error=True)
+        self.br.add_protocols('OpenFlow13', 'OpenFlow14')
+        self.assertEqual(
+            self.br.db_get_val('Bridge', self.br.br_name, 'protocols'),
+            ['OpenFlow10', 'OpenFlow12', 'OpenFlow13', 'OpenFlow14'])
+
+    def test_add_protocols_add_existing(self):
+        self.br.set_db_attribute('Bridge', self.br.br_name, 'protocols',
+                                 ['OpenFlow10', 'OpenFlow12', 'OpenFlow13'],
+                                 check_error=True)
+        self.br.add_protocols('OpenFlow13')
+        self.assertEqual(
+            self.br.db_get_val('Bridge', self.br.br_name, 'protocols'),
+            ['OpenFlow10', 'OpenFlow12', 'OpenFlow13'])
+
+    def _test_set_igmp_snooping_state(self, state):
+        self.br.set_igmp_snooping_state(state)
+        self.assertEqual(
+            state,
+            self.br.db_get_val(
+                'Bridge', self.br.br_name, 'mcast_snooping_enable'))
+        br_other_config = self.ovs.ovsdb.db_find(
+            'Bridge', ('name', '=', self.br.br_name), columns=['other_config']
+        ).execute()[0]['other_config']
+        self.assertEqual(
+            str(state),
+            br_other_config['mcast-snooping-disable-flood-unregistered'])
+
+    def test_set_igmp_snooping_enabled(self):
+        self._test_set_igmp_snooping_state(True)
+
+    def test_set_igmp_snooping_disabled(self):
+        self._test_set_igmp_snooping_state(False)
 
     def test_get_datapath_id(self):
         brdev = ip_lib.IPDevice(self.br.br_name)
@@ -159,8 +223,8 @@ class OVSBridgeTestCase(OVSBridgeTestBase):
 
     def test_add_tunnel_port_ipv4(self):
         attrs = {
-            'remote_ip': '192.0.2.1',  # RFC 5737 TEST-NET-1
-            'local_ip': '198.51.100.1',  # RFC 5737 TEST-NET-2
+            'remote_ip': self.get_test_net_address(1),
+            'local_ip': self.get_test_net_address(2),
         }
         self._test_add_tunnel_port(attrs)
 
@@ -171,6 +235,32 @@ class OVSBridgeTestCase(OVSBridgeTestBase):
         }
         self._test_add_tunnel_port(attrs)
 
+    def test_add_tunnel_port_custom_port(self):
+        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
+        self.br.add_tunnel_port(
+            port_name,
+            self.get_test_net_address(1),
+            self.get_test_net_address(2),
+            tunnel_type=const.TYPE_VXLAN,
+            vxlan_udp_port=12345)
+        options = self.ovs.db_get_val('Interface', port_name, 'options')
+        self.assertEqual("12345", options['dst_port'])
+
+    def test_add_tunnel_port_tos(self):
+        attrs = {
+            'remote_ip': self.get_test_net_address(1),
+            'local_ip': self.get_test_net_address(2),
+            'tos': 'inherit',
+        }
+        port_name = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
+        self.br.add_tunnel_port(port_name, attrs['remote_ip'],
+                                attrs['local_ip'], tos=attrs['tos'])
+        self.assertEqual('gre',
+                         self.ovs.db_get_val('Interface', port_name, 'type'))
+        options = self.ovs.db_get_val('Interface', port_name, 'options')
+        for attr, val in attrs.items():
+            self.assertEqual(val, options[attr])
+
     def test_add_patch_port(self):
         local = utils.get_rand_device_name(net_helpers.PORT_PREFIX)
         peer = 'remotepeer'
@@ -334,6 +424,145 @@ class OVSBridgeTestCase(OVSBridgeTestBase):
         self.assertIsNone(max_rate)
         self.assertIsNone(burst)
 
+    def test_ingress_bw_limit(self):
+        port_name, _ = self.create_ovs_port()
+        self.br.update_ingress_bw_limit_for_port(port_name, 700, 70)
+        max_rate, burst = self.br.get_ingress_bw_limit_for_port(port_name)
+        self.assertEqual(700, max_rate)
+        self.assertEqual(70, burst)
+
+        self.br.update_ingress_bw_limit_for_port(port_name, 750, 100)
+        max_rate, burst = self.br.get_ingress_bw_limit_for_port(port_name)
+        self.assertEqual(750, max_rate)
+        self.assertEqual(100, burst)
+
+        self.br.delete_ingress_bw_limit_for_port(port_name)
+        max_rate, burst = self.br.get_ingress_bw_limit_for_port(port_name)
+        self.assertIsNone(max_rate)
+        self.assertIsNone(burst)
+
+    def test_ingress_bw_limit_dpdk_port(self):
+        port_name, _ = self.create_ovs_port(
+            ('type', agent_const.OVS_DPDK_VHOST_USER))
+        self.br.update_ingress_bw_limit_for_port(port_name, 700, 70)
+        max_rate, burst = self.br.get_ingress_bw_limit_for_dpdk_port(
+            port_name)
+        self.assertEqual(700, max_rate)
+        self.assertEqual(70, burst)
+
+        self.br.update_ingress_bw_limit_for_port(port_name, 750, 100)
+        max_rate, burst = self.br.get_ingress_bw_limit_for_dpdk_port(
+            port_name)
+        self.assertEqual(750, max_rate)
+        self.assertEqual(100, burst)
+
+        self.br.delete_ingress_bw_limit_for_port(port_name)
+        max_rate, burst = self.br.get_ingress_bw_limit_for_dpdk_port(
+            port_name)
+        self.assertIsNone(max_rate)
+        self.assertIsNone(burst)
+
+    def test_db_create_references(self):
+        with self.ovs.ovsdb.transaction(check_error=True) as txn:
+            queue = txn.add(self.ovs.ovsdb.db_create("Queue",
+                                                     other_config={'a': '1'}))
+            qos = txn.add(self.ovs.ovsdb.db_create("QoS", queues={0: queue}))
+            txn.add(self.ovs.ovsdb.db_set("Port", self.br.br_name,
+                                          ('qos', qos)))
+
+        def cleanup():
+            with self.ovs.ovsdb.transaction() as t:
+                t.add(self.ovs.ovsdb.db_destroy("QoS", qos.result))
+                t.add(self.ovs.ovsdb.db_destroy("Queue", queue.result))
+                t.add(self.ovs.ovsdb.db_clear("Port", self.br.br_name, 'qos'))
+
+        self.addCleanup(cleanup)
+        val = self.ovs.ovsdb.db_get("Port", self.br.br_name, 'qos').execute()
+        self.assertEqual(qos.result, val)
+
+    def test_db_add_set(self):
+        protocols = ["OpenFlow10", "OpenFlow11"]
+        expected = self.br.initial_protocols.union(protocols)
+        self.br.ovsdb.db_add("Bridge", self.br.br_name, "protocols",
+                             *protocols).execute(check_error=True)
+        self.assertItemsEqual(expected,
+                              self.br.db_get_val('Bridge',
+                                                 self.br.br_name, "protocols"))
+
+    def test_db_add_map(self):
+        key = "testdata"
+        data = {key: "testvalue"}
+        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
+                             data).execute(check_error=True)
+        self.assertEqual(data, self.br.db_get_val('Bridge', self.br.br_name,
+                                                  'external_ids'))
+        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
+                             {key: "newdata"}).execute(check_error=True)
+        self.assertEqual(data, self.br.db_get_val('Bridge', self.br.br_name,
+                                                  'external_ids'))
+
+    def test_db_add_map_multiple_one_dict(self):
+        data = {"one": "1", "two": "2", "three": "3"}
+        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
+                             data).execute(check_error=True)
+        self.assertEqual(data, self.br.db_get_val('Bridge', self.br.br_name,
+                                                  'external_ids'))
+
+    def test_db_add_map_multiple_dicts(self):
+        data = ({"one": "1"}, {"two": "2"}, {"three": "3"})
+        self.br.ovsdb.db_add("Bridge", self.br.br_name, "external_ids",
+                             *data).execute(check_error=True)
+        combined = {k: v for a in data for k, v in a.items()}
+        self.assertEqual(combined, self.br.db_get_val('Bridge',
+                                                      self.br.br_name,
+                                                      'external_ids'))
+
+    def test_db_add_ref(self):
+        ovsdb = self.ovs.ovsdb
+        brname = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
+        br = ovs_lib.OVSBridge(brname)  # doesn't create
+        self.addCleanup(br.destroy)
+        with ovsdb.transaction(check_error=True) as txn:
+            br = txn.add(ovsdb.db_create('Bridge', name=brname))
+            txn.add(ovsdb.db_add('Open_vSwitch', '.', 'bridges', br))
+
+        self.assertIn(brname, self.ovs.get_bridges())
+
+    def test_db_add_to_new_object(self):
+        ovsdb = self.ovs.ovsdb
+        brname = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
+        br = ovs_lib.OVSBridge(brname)  # doesn't create
+        self.addCleanup(br.destroy)
+        with ovsdb.transaction(check_error=True) as txn:
+            txn.add(ovsdb.add_br(brname))
+            txn.add(ovsdb.db_add('Bridge', brname, 'protocols', 'OpenFlow10'))
+
+    def test_cascading_del_in_txn(self):
+        ovsdb = self.ovs.ovsdb
+        port_name, _ = self.create_ovs_port()
+
+        def del_port_mod_iface():
+            with ovsdb.transaction(check_error=True) as txn:
+                txn.add(ovsdb.del_port(port_name, self.br.br_name,
+                                       if_exists=False))
+                txn.add(ovsdb.db_set('Interface', port_name,
+                                     ('type', 'internal')))
+        self.assertRaises((RuntimeError, idlutils.RowNotFound),
+                          del_port_mod_iface)
+
+    def test_delete_flows_all(self):
+        self.br.add_flow(in_port=1, actions="output:2")
+        self.br.delete_flows(cookie=ovs_lib.COOKIE_ANY)
+        self.assertEqual([], self.br.dump_all_flows())
+
+    def test_delete_flows_strict(self):
+        self.br.delete_flows(cookie=ovs_lib.COOKIE_ANY)  # remove NORMAL action
+        self.br.add_flow(in_port=1, actions="output:2")
+        self.br.add_flow(in_port=1, priority=100, actions="output:3")
+        self.assertEqual(2, len(self.br.dump_all_flows()))
+        self.br.delete_flows(in_port=1, priority=100, strict=True)
+        self.assertEqual(1, len(self.br.dump_all_flows()))
+
 
 class OVSLibTestCase(base.BaseOVSLinuxTestCase):
 
@@ -358,6 +587,9 @@ class OVSLibTestCase(base.BaseOVSLinuxTestCase):
 
     def test_bridge_lifecycle_ovsbridge(self):
         name = utils.get_rand_name(prefix=net_helpers.BR_PREFIX)
+        mac_table_size = 12345
+        cfg.CONF.set_override(
+            'bridge_mac_table_size', mac_table_size, group='OVS')
         br = ovs_lib.OVSBridge(name)
         self.assertEqual(br.br_name, name)
         # Make sure that instantiating an OVSBridge does not actually create
@@ -365,6 +597,11 @@ class OVSLibTestCase(base.BaseOVSLinuxTestCase):
         self.addCleanup(self.ovs.delete_bridge, name)
         br.create()
         self.assertTrue(self.ovs.bridge_exists(name))
+        br_other_config = self.ovs.ovsdb.db_find(
+            'Bridge', ('name', '=', name), columns=['other_config']
+        ).execute()[0]['other_config']
+        self.assertEqual(str(mac_table_size),
+                         br_other_config['mac-table-size'])
         br.destroy()
         self.assertFalse(self.ovs.bridge_exists(name))
 
